"""The :mod:`fitness` module is intended to contain parent classes to create
the procedures required to evaluate the experiment phenotypes for evolution.

Similarly to the rest of modules in this library, a class desing is selected
so the user can define multiple methods to obtain a highly customisable
fitness object. Similar aproaches are commonly used in python, such for
example Unittests.

A rework form:
https://github.com/DEAP/deap/blob/master/deap/base.py
"""

from abc import ABC, abstractmethod

from dask import compute, delayed

from gevopy import genetics


class _HasCache():
    """Extend class Fitness with cache properties."""

    def __init__(self, *args, cache: bool = False, **kwargs) -> None:
        """Generic constructor for fitness objects.
        :param cache: Enables cache with True, default is False
        """
        super().__init__(*args, **kwargs)
        self.cache = cache

    @property
    def cache(self):
        """Returns a boolean value indicating if score cache is enabled.
        :return: True if score cache is enabled, otherwise False
        """
        return self._cv is not None

    @cache.setter
    def cache(self, value):
        """Enables or disables cache on the fitness object.
        :param value: True to enable cache, otherwise False
        """
        if not isinstance(value, bool):
            raise ValueError("Expected 'bool' type for cache")
        self._cv = {} if value else None


class _HasScheduler():
    """Extend class Fitness dask scheduler properties."""
    scheduler_options = ['synchronous', 'threads', 'processes']

    def __init__(self, *args, scheduler: str = "threads", **kwargs) -> None:
        """Generic constructor for fitness objects.
        :param scheduler: Dask scheduler to use during fitness evaluation
        """
        super().__init__(*args, **kwargs)
        self.scheduler = scheduler

    @property
    def scheduler(self):
        """Returns the configured value for dask scheduler in the object.
        :return: A string containing one of the 'scheduler_options'
        """
        return self.__scheduler

    @scheduler.setter
    def scheduler(self, value):
        """Configures/edits the dask scheduler for phenotypes evaluation.
        :param value: A string containing one of the 'scheduler_options'
        """
        if value not in self.scheduler_options:
            raise ValueError(f"Unknown '{value}' value for scheduler")
        self.__scheduler = value


class FitnessModel(_HasCache, _HasScheduler, ABC):
    """Fitness base class to run phenotypes evaluation.
    It requires the user to define a `score` method which takes a phenotype
    as input and returns its score value. Use cache and scheduler to control
    how the evaluation is executed:

      - If phenotype score does not change between generations, you can use
        `cache=True` to skip score computation on those phenotypes whose id
        has been already evaluated.

      - If phenotypes interact between them, or have some waiting times on
        the evaluation process, you might want to set `scheduler="threads"`
        to run multiple phenotypes in parallel. Scheduler is based on Dask,
        see https://docs.dask.org/en/stable/scheduling.html

    Additionally you can configure a setUp function which would be executed
    once, before evaluating the phenotypes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, phenotypes):
        """Fitness call to calculate phenotype scores.
        :param phenotypes: List with phenotypes to score
        """
        self.setup(phenotypes)
        delayeds = [delayed(self.worker)(ph) for ph in phenotypes]
        scores = compute(*delayeds, scheduler=self.scheduler)
        for phenotype, score in zip(phenotypes, scores):
            phenotype.score = score
        if self.cache:
            self._cv = {ph.id: ph.score for ph in phenotypes}

    def setup(self, phenotypes):
        """Fitness function designed to prepare the phenotypes evaluation.
        :param phenotypes: List with phenotypes to score
        """
        pass  # pylint: disable=unnecessary-pass

    def worker(self, phenotype):
        """Fitness wrap to return cached score value if configured.
        :param phenotype: Phenotype to evaluate
        :return: Phenotype score
        """
        assert isinstance(phenotype, genetics.GenotypeModel)
        if self.cache and phenotype.id in self._cv:
            return self._cv[phenotype.id]
        else:
            return self.score(phenotype)

    @abstractmethod
    def score(self, phenotype):
        """Abstract function to return the score value of a phenotype.
        :param phenotype: Phenotype to evaluate
        :return: Phenotype score
        """
        raise NotImplementedError

    @classmethod
    def __get_validators__(cls):
        yield cls.class_validator

    @classmethod
    def class_validator(cls, value):
        """Validates the value is a correct FitnessModel type."""
        if not isinstance(value, cls):
            raise TypeError("'FitnessModel' type required")
        return value



"""The :mod:`selection` module is intended to contain abstract and specific
classes to design your evolution algoritms.

Selection is the stage of a genetic algorithm in which individual genomes
are chosen from a population for later breeding (using the crossover operator).

A rework form:
https://github.com/DEAP/deap/blob/master/deap/tools/selection.py
"""
# pylint: disable=too-few-public-methods

import builtins
import inspect
import itertools
import math
import random
import types
from abc import ABC, abstractmethod

import numpy as np

from . import Pool

methods = {"Ponderated", "Uniform", "Best", "Worst", "Tournaments"}
__all__ = methods.union({"Selection"})


class Selection(ABC):
    """A generic selection objext that returns an specific amount of phenotypes
    from a pool according to their scores. The main objective of this
    selection procedure is to keep, crossover or mutate only those phenotypes
    that perform better on the solution of a problem or environment.
    """

    @abstractmethod
    def __call__(self, pool, n):
        """Executes the selection.
        :param pool: A Pool of phenotypes.
        :param n: The number of phenotypes to select
        :returns: List with **n** selected phenotypes
        """
        raise NotImplementedError

    @classmethod
    def __get_validators__(cls):
        yield cls.class_validator

    @classmethod
    def class_validator(cls, value):
        """Validates the value is a correct Selection type."""
        if not isinstance(value, cls):
            raise TypeError("'Selection' type required")
        return value


class Ponderated(Selection):
    """Select *n* random phenotypes where each phenotype probability is
    ponderated by the normalized score (score[i]/sum(scores)). The list
    returned contains references to the *pool phenotypes*.
    """

    def __call__(self, pool, n):
        """Executes the selection of 'n' phenotypes from a pool.
        :param pool: A Pool of phenotypes.
        :param n: The number of phenotypes to select
        :returns: A list of selected phenotypes.
        """
        match (pool, n):
            case _ if not isinstance(pool, Pool):
                raise ValueError("Expected type 'Pool' for 'pool'")
            case _ if not isinstance(n, int):
                raise ValueError("Expected type 'int' for 'n'")
            case _ if not n >= 0:
                raise ValueError("Value for 'n' cannot be lower than 0")

        sum_scores = sum(pool.scores())
        try:  # Divide R[0,1] segment with normalised phenotypes weigths
            n_scores = [score / sum_scores for score in pool.scores()]
            n_prb = list(itertools.accumulate(n_scores))
        except ZeroDivisionError:
            return Uniform.__call__(self, pool, n)

        def choose(p):  # next avoids generating the whole list
            return next(x for x, r in zip(pool, n_prb) if r > p)
        return [choose(p) for p in np.random.rand(n)]


class Uniform(Selection):
    """Select *n* random phenotypes where each phenotype probability is
    equal. The list returned contains references to the *pool phenotypes*.
    """

    def __call__(self, pool, n):
        """Executes the selection of 'n' phenotypes from a pool.
        :param pool: A Pool of phenotypes.
        :param n: The number of phenotypes to select
        :returns: A list of selected phenotypes.
        """
        match (pool, n):
            case _ if not isinstance(pool, Pool):
                raise ValueError("Expected type 'Pool' for 'pool'")
            case _ if not isinstance(n, int):
                raise ValueError("Expected type 'int' for 'n'")
            case _ if not n >= 0:
                raise ValueError("Value for 'n' cannot be lower than 0")

        return [random.choice(pool) for _ in range(n)]


class Best(Selection):
    """Returns the best phenotype among the input *phenotypes* `n` times.
    The returned list contains references to the input *phenotype*.
    """

    def __call__(self, pool, n):
        """Executes the selection of 'n' phenotypes from a pool.
        :param pool: A Pool of phenotypes.
        :param n: The number of phenotypes to return
        :returns: A list of selected phenotypes.
        """
        match (pool, n):
            case _ if not isinstance(pool, Pool):
                raise ValueError("Expected type 'Pool' for 'pool'")
            case _ if not isinstance(n, int):
                raise ValueError("Expected type 'int' for 'n'")
            case _ if not n >= 0:
                raise ValueError("Value for 'n' cannot be lower than 0")

        return [pool[0] for _ in range(n)]


class Worst(Selection):
    """Returns the worst phenotype among the input *phenotypes* `n` times.
    The returned list contains references to the input *phenotype*.
    """

    def __call__(self, pool, n):
        """Executes the selection of 'n' phenotypes from a pool.
        :param pool: A Pool of phenotypes.
        :param n: The number of phenotypes to return
        :returns: A list of selected phenotypes.
        """
        match (pool, n):
            case _ if not isinstance(pool, Pool):
                raise ValueError("Expected type 'Pool' for 'pool'")
            case _ if not isinstance(n, int):
                raise ValueError("Expected type 'int' for 'n'")
            case _ if not n >= 0:
                raise ValueError("Value for 'n' cannot be lower than 0")

        return [pool[-1] for _ in range(n)]


class Tournaments(Selection):
    """Select the best phenotype among *tournsize* randomly chosen phenotypes,
    *n* times. The list returned contains references to the input *phenotypes*.
    """

    def __init__(self, tournsize=lambda n: math.floor(math.sqrt(n))):
        """Selection constructor for tournaments. It takes an optional
        parameter **tournsize** which defines or calculates the size of
        each  tournament participants.
        :param tournsize: Positive number or function, default is the square
        root floor of **n**.
        """
        match type(tournsize):
            case builtins.int if tournsize <= 1:
                raise ValueError("'tournsize' cannot be lower than 1")
            case builtins.int:
                self.tournsize = lambda _: tournsize
            case types.LambdaType:
                if len(inspect.signature(tournsize).parameters) != 1:
                    raise ValueError("Arity for 'tournsize' must be 1")
                self.tournsize = tournsize
            case _:
                raise ValueError("'tournsize' must be 'int' or 'LambdaType'")

    def __call__(self, pool, n):
        """Executes the selection of 'n' phenotypes from a pool.
        :param pool: A Pool of phenotypes.
        :param n: The number of phenotypes to select
        :returns: A list of selected phenotypes.
        """
        match (pool, n):
            case _ if not isinstance(pool, Pool):
                raise ValueError("Expected type 'Pool' for 'pool'")
            case _ if not isinstance(n, int):
                raise ValueError("Expected type 'int' for 'n'")
            case _ if not n >= 0:
                raise ValueError("Value for 'n' cannot be lower than 0")

        tournsize = self.tournsize(n)
        aspirants = [Uniform.__call__(self, pool, tournsize) for _ in range(n)]
        return [max(x, key=lambda x: x.score) for x in aspirants]



"""The :mod:`mutation` module is intended to contain abstract and specific
classes to design your evolution algoritms.

Mutation is the stage of a genetic algorithm in which individual genomes
from an outspring are changed in order to generate new features.

https://github.com/DEAP/deap/blob/master/deap/tools/mutation.py
"""
# pylint: disable=too-few-public-methods

from abc import ABC, abstractmethod
from copy import deepcopy

import numpy as np

from gevopy import genetics

methods = {"SinglePoint"}
__all__ = methods.union({"Mutation"})


class Mutation(ABC):
    """Mutation is a genetic operator used to maintain genetic diversity
    from one generation of a population of genetic algorithm chromosomes
    to the next. It is analogous to biological mutation.

    A mutation operator involves a probability that an arbitrary bit in
    a genetic chain will be flipped from its original state.

    Mutation operators are used in an attempt to avoid local minima by
    preventing the population of chromosomes from becoming too similar
    to each other, thus slowing or even stopping convergence to the global
    optimum. This reasoning also leads most GA systems to avoid only taking
    the fittest of the population in generating the next generation, but
    rather selecting a random (or semi-random) set with a weighting toward
    those that are fitter.

    For different genome types, different mutation types are suitable.
    """

    def __call__(self, phenotype):
        """Executes the mutation on a phenotype. If the phenotype was never
        evaluated, then clone is an exact copy maintaining id and parents.
        :param phenotype: The phenotype to mutate
        :return: Mutated phenotype
        """
        match phenotype:
            case _ if not isinstance(phenotype, genetics.GenotypeModel):
                raise ValueError("Expected 'GenotypeModel' for 'phenotype'")

        phenotype_copy = deepcopy(phenotype)
        self.mutate_features(phenotype_copy.__dict__.values())
        return phenotype_copy

    def mutate_features(self, features):
        """Recursively mutates phenotype features. For example a list of
        diploids in the case of Eucaryote genotype.
        :param features: Phenotype list of feature values
        """
        for value in features:
            if isinstance(value, genetics.Chromosome):
                self.mutate_chromosome(chromosome=value)
            elif isinstance(value, list):
                self.mutate_features(features=value)

    @abstractmethod
    def mutate_chromosome(self, chromosome):
        """Executes mutation on a specific object using a probability mask.
        :param chromosome: The value chain where to apply mutation
        :param mask: Boolean array indicating where to apply mutation
        """
        raise NotImplementedError

    @classmethod
    def __get_validators__(cls):
        yield cls.class_validator

    @classmethod
    def class_validator(cls, value):
        """Validates the value is a correct Mutation type."""
        if not isinstance(value, cls):
            raise TypeError("'Mutation' type required")
        return value


class SinglePoint(Mutation):
    """Executes single point mutation over a property array. This mutation
    method generates a random variable for each bit in a chain sequence to
    evaluate whether or not a particular bit will be flipped.
    For Example:
    - chromosome.chain => [0110101011101011000101010010]
                            ↓    ↓↓↓ ↓    ↓ ↓    ↓ ↓↓↓
    - chromosome.chain => [0010100101001010010101111100]

    The chromosome chain is modified in place and keep hist original length.
    This mutation procedure is based on the biological point mutation.
    """

    def __init__(self, mutpb=0.05):
        """Mutation generic constructor.
        :param mutpb: Probability of each slot mutation, defaults to 0.05
        """
        self.mutation_probability = mutpb

    def mutate_chromosome(self, chromosome):
        """Executes the mutation. The chromosome chain is modified in place.
        :param chromosome: Chromosome to be mutated
        """
        mask = np.random.random(chromosome.size) <= self.mutation_probability
        chromosome[mask] = chromosome[mask].__mutate__()



"""Module to test database operations for evolution algorithms."""
# pylint: disable=redefined-outer-name

import uuid

from pytest import fixture


@fixture(scope="module", autouse=True)
def session(db_interface):
    """Fixture to start a graph database session for testing"""
    with db_interface.session() as session:
        yield session


@fixture(scope="function")
def experiment_name():
    """Fixture to generate an experiment name for testing"""
    return f"Experiment_{uuid.uuid4()}"


@fixture(scope="function", autouse=True)
def clean_experiment(session, experiment_name):
    """Fixture to start a graph database session for testing"""
    yield None
    session.del_experiment(experiment_name)


@fixture(scope="function", autouse=True)
def add_experiment(population, experiment_name):
    """Fixture to add experiment to some phenotypes"""
    for phenotype in population[::2]:
        phenotype.experiment = experiment_name


@fixture(scope="function", autouse=True)
def add_parents(population):
    """Fixture to create parent relatioships between phenotypes"""
    for child, parent in zip(population, population[1:]):
        child.parents = [parent.id]


@fixture(scope="function")
def phenotypes(population):
    """Fixture to return phenotypes ids from population"""
    return [p.dict(serialize=True) for p in population]


@fixture(scope="function")
def ids(population):
    """Fixture to return phenotypes ids from population"""
    return [str(ph.id) for ph in population]


def test_rwd_phenotypes(session, phenotypes, ids):
    """Test write, read and delete of phenotypes in db"""
    assert ids == session.add_phenotypes(phenotypes)
    assert phenotypes == session.get_phenotypes(ids)
    assert ids == session.del_phenotypes(ids)



"""Module with utilities for the evolution modules"""

import numpy as np


def cross_chromosomes(chromosome_1, chromosome_2, points):
    """Crosses chromosome arrays on the specified points.
    :param chromosome_1: First array to cross
    :param chromosome_2: Second array to cross
    :param points: List of points to cross
    """
    ch1, ch2 = chromosome_1, chromosome_2  # Short code
    ax1, ax2 = np.split(ch1, points), np.split(ch2, points)
    ax1[1::2], ax2[1::2] = ax2[1::2], ax1[1::2]
    ch1[:], ch2[:] = np.concatenate(ax1), np.concatenate(ax2)



"""Module to test crossover operations for evolution algorithms."""
# pylint: disable=redefined-outer-name

from inspect import ismethod, signature

from pytest import fixture, mark

from gevopy import genetics
from gevopy.tools import crossover


# Module fixtures ---------------------------------------------------
@fixture(scope="class")
def children(crossover, parents):
    """Fixture to return crossovered phenotypes from base"""
    return crossover(parents[0], parents[1])


@fixture(scope="class")
def parents(genotype):
    """Fixture to generate a base phenotypes to cross"""
    return genotype(), genotype()


@fixture(scope="class")
def generation(parents):
    """Fixture to calculate the parents generation"""
    return max([ph.generation for ph in parents])


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Crossover instances attributes"""

    def test_is_callable(self, crossover):
        """Test crossover is a callable with arity 2"""
        assert callable(crossover)
        assert len(signature(crossover).parameters) == 2

    def test_attr_cross_chromosomes(self, crossover):
        """Test attribute cross_chromosome is a method"""
        assert hasattr(crossover, "cross_chromosomes")
        assert ismethod(crossover.cross_chromosomes)


class ExecutionRequirements:
    """Tests group for Crossover execution features"""

    def test_keeps_genotype(self, parents, children):
        """Test crossover does not alter phenotype"""
        assert all(isinstance(x, genetics.GenotypeModel) for x in children)
        assert all(type(p) is type(x) for p, x in zip(parents, children))

    def test_id_change(self, parents, children):
        """Test ids on offsprings are different from parents"""
        assert children[0].id not in [p.id for p in parents]
        assert children[1].id not in [p.id for p in parents]

    def test_parents(self, parents, children):
        """Test children parents attr is correctly set"""
        assert set(children[0].parents) == set(ph.id for ph in parents)
        assert set(children[1].parents) == set(ph.id for ph in parents)

    def test_generation(self, generation, children):
        """Test children generation attr is correctly set"""
        assert children[0].generation == generation + 1
        assert children[1].generation == generation + 1


# Parametrization ---------------------------------------------------
class TestOnePoint(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'OnePoint' crossover"""

    @fixture(scope="class")
    def crossover(self):
        """Parametrization to define the crossover method to use"""
        return crossover.OnePoint()

    def test_chromosome(self, parents, children):
        """Test chromosomes are different from parents"""
        assert parents[0].chromosome != children[0].chromosome
        assert parents[1].chromosome != children[1].chromosome


class TestTwoPoint(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'TwoPoint' crossover"""

    @fixture(scope="class")
    def crossover(self):
        """Parametrization to define the crossover method to use"""
        return crossover.TwoPoint()

    def test_chromosome(self, parents, children):
        """Test chromosomes are different from parents"""
        assert parents[0].chromosome != children[0].chromosome
        assert parents[1].chromosome != children[1].chromosome


class TestMultiPoint(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'MultiPoint' crossover"""

    @fixture(scope="class", params=[1, 2, 5])
    def crossover(self, request):
        """Parametrization to define the crossover method to use"""
        return crossover.MultiPoint(n=request.param)

    def test_chromosome(self, parents, children):
        """Test chromosomes are different from parents"""
        assert parents[0].chromosome != children[0].chromosome
        assert parents[1].chromosome != children[1].chromosome


class TestUniform(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Uniform' crossover"""

    @fixture(scope="class", params=[0.1, 0.2, 0.5])
    def crossover(self, request):
        """Parametrization to define the crossover method to use"""
        return crossover.Uniform(indpb=request.param)

    @mark.parametrize("crossover", [0.0], indirect=True)
    def test_indpb_000(self, parents, children):
        """Tests that no crossover does not modity chromosome"""
        assert parents[0].chromosome == children[0].chromosome
        assert parents[1].chromosome == children[1].chromosome

    @mark.parametrize("crossover", [1.0], indirect=True)
    def test_indpb_100(self, parents, children):
        """Tests that crossover does modity chromosome"""
        assert parents[0].chromosome != children[0].chromosome
        assert parents[1].chromosome != children[1].chromosome



"""The :mod:`genetics` module is intended to contain generic and specific
classes to design your Genotype models.

Define your Genotypes following the `dataclass` principles from `pydantic` by
using the base model `GenotypeModel`. All dataclass attributes are accepted in
addition to an special type `Chromosome` provided in the module. To start
use the already defined chromosome subclasses such `Haploid` and `Diploid`
depending on the complexity of your genetic model.
"""

import copy
import json
import uuid
from datetime import datetime
from pprint import pformat
from typing import List, MutableSequence

import numpy as np
from pydantic import BaseModel, Field, PositiveInt

from gevopy import random as ea_random


class Chromosome(np.ndarray, MutableSequence):
    """A chromosome is a long DNA molecule with part or all of the genetic
    material of an organism. In the case of Evolutionary Algorithms it
    contains the information required to evaluate a phenotype.

    This library bases chromosomes on numpy.ndarray to and therefore when
    creating one, the steps defined at `subclassing ndarray` must be followed.
    Additionally it subclasses from `collections.abc.MutableSequence` to
    provide additional and standard python methods such as `count`.

    In order to suport serialization/deserialization the new method must
    accept only the first input parameter for the function np.array.
    """

    def __new__(cls, data):
        """Standard python method required to subclass np.ndarray.
        :param cls: Chromosome ndarray subclass type
        :param data: Any object exposing the array interface or sequence
        :param pairs: Number of pairs the chromosome represents
        :return: A new generated Chromosome instance
        """
        return np.array(data, dtype="uint8").view(cls)

    def __array_finalize__(self, obj):
        """Mechanism that numpy provides to allow subclasses to handle
        the various ways that new instances get created.
        :param obj: New instance provided for slicing and `view`
        """
        if obj is None:
            return

    def insert(self, index, value):
        """Method insert is unsupported on chromosome types"""
        raise AttributeError(f"Unsupported operation by '{self.__class__}'")

    def __delitem__(self, index):
        """Method delitem is unsupported on chromosome types"""
        raise AttributeError(f"Unsupported operation by '{self.__class__}'")

    def __mutate__(self):
        """Performs the chromosome mutation operation.
        :return: Chromosome with mutated values
        """
        raise NotImplementedError

    def __eq__(self, other):
        """Standard python method to compares 2 chromosomes equality.
        :param other: Chromosome to compare with
        :return: Boolean, True if are equal, otherwise False
        """
        return np.array_equal(self, other)

    def __ne__(self, other):
        """Standard python method to compares 2 chromosomes non equality.
        :param other: Chromosome to compare with
        :return: Boolean, True if are different, otherwise False
        """
        return ~self.__eq__(other)

    def __cross__(self, other):
        """Magic method to compare a chromosome bitwise.
        :param other: Chromosome to compare with
        :return: List of chromosome bits where equal
        """
        return super().__eq__(other)

    @classmethod
    def __get_validators__(cls):
        """Pydantic magic method for custom validation and deserialization.
        :yield: Validation method for construction
        """
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        """Validation method for construction and deserialization.
        :param val: Value to pass to __new__ data
        :return: Deserialized chromosome
        """
        return cls(data=val)


class Haploid(Chromosome):
    """The word haploid describes a condition, a cell, or an organism that
    contains half of the set of homologous chromosomes present in the somatic
    cell. Homologous chromosomes are two chromosomes that pair up by having
    the same gene sequence, loci, chromosomal length, and centromere location.

    Half of the homologous pairs are maternal (coming from the mother) whereas
    the other half, paternal (coming from the father). Thus, in other words,
    a haploid is when a cell, for instance, contains half of the total
    homologous chromosomes, i.e. a single set of chromosomes that are unpaired.
    """

    @classmethod
    @property
    def states(cls):
        """Returns the number of possible chromosome states.
        :return: Chromosome with inverted values
        """
        return 2

    def __invert__(self):
        """Computes and returns the bit-wise inversion of the chromosome.
        :return: Chromosome with inverted values
        """
        return np.logical_not(self).view("uint8")

    def __mutate__(self):
        """Computes and returns the bit-wise mutation of the chromosome.
        :return: Chromosome with mutated values
        """
        return ea_random.haploid(self.size)


class Diploid(Chromosome):
    """In genetics and biology, the term diploid refers to the cell containing
    two sets of homologous chromosomes wherein each chromosome in a set is
    obtained from each of the two-parent cells.

    As example, the fusion of two haploid sex cells results in the formation
    of a diploid cell called a zygote.
    """

    @classmethod
    @property
    def states(cls):
        """Returns the number of possible chromosome states.
        :return: Chromosome with inverted values
        """
        return 4

    def __invert__(self):
        """Computes and returns the bit-wise inversion of the chromosome.
        :return: Chromosome with inverted values
        """
        return np.bitwise_and(super().__invert__(), 3)

    def __mutate__(self):
        """Computes and returns the bit-wise mutation of the chromosome.
        :return: Chromosome with mutated values
        """
        return ea_random.diploid(self.size)


class Triploid(Chromosome):
    """Similar to haploid and diploid, cells can contain three sets of
    homologous chromosomes, increasing the amount of combinations for each
    bit (trinary) position.

    In biology triploidy is a rare chromosomal abnormality.
    """

    @classmethod
    @property
    def states(cls):
        """Returns the number of possible chromosome states.
        :return: Chromosome with inverted values
        """
        return 8

    def __invert__(self):
        """Computes and returns the bit-wise inversion of the chromosome.
        :return: Chromosome with inverted values
        """
        return np.bitwise_and(super().__invert__(), 7)

    def __mutate__(self):
        """Computes and returns the bit-wise mutation of the chromosome.
        :return: Chromosome with mutated values
        """
        return ea_random.triploid(self.size)


class GenotypeModel(BaseModel):
    """Evolution Genotype is the most basic but flexible form of genetics.
    It is a chromosomes container with an unique identifier. Different
    organisms might have different numbers of chromosomes.

    When subclassing a genotype note the following attributes are reserved:
     - id: A unique identifier for the phenotype
     - experiment: Name of the experiment the phenotype belongs, can be None
     - created: Datetime when the phenotype was instantiated
     - parents: List of phenotype ids used to generate the phenotype
     - generation: Positive integer indicating the evolution generations
     - score: Float indicating the phenotype score (None, when not evaluated)
     - clone: Method to produce a genotype deep copy with different id
    """

    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    experiment: str = None
    created: datetime = Field(default_factory=datetime.utcnow)
    parents: List[uuid.UUID] = []
    generation: PositiveInt = Field(default=1)
    score: float = None

    class Config:
        # pylint: disable=missing-class-docstring
        # pylint: disable=too-few-public-methods
        json_encoders = {Chromosome: lambda x: x.astype("uint8").tolist()}

    def __repr__(self):
        return pformat(self.__dict__, sort_dicts=False)

    def dict(self, *args, serialize=False, **kwargs):
        """Returns the phenotype serialized and in dict form,
        :return: Serialized dictionary
        """
        # TODO: This is really slow, improve round trip, see
        # https://github.com/pydantic/pydantic/issues/1409#issuecomment-650116116
        if serialize:
            return json.loads(self.json())
        return super().dict(*args, **kwargs)

    def clone(self):
        """Copies the phenotype using a different id and empty score.
        :return: Phenotype copy
        """
        clone = copy.deepcopy(self)
        clone.id = uuid.uuid4()  # Generate new id
        clone.score = None  # Reset the clone score
        return clone



"""Module for evolution requirements fixtures"""
# pylint: disable=redefined-outer-name

import os

from neo4j import GraphDatabase
from pytest import fixture
from gevopy.database import Neo4jInterface, EmptyInterface

import examples.algorimths
import examples.evaluation
import examples.genotypes

NEO4J_DRIVER = dict(
    uri=os.getenv('TESTING_NEO4J_URI'),
    auth=os.getenv('TESTING_NEO4J_AUTH'),
)


@fixture(scope="session", params=[NEO4J_DRIVER])
def driver_kwds(request):
    """Fixture to generate a neo4j database driver"""
    # pylint: disable=not-context-manager
    with GraphDatabase.driver(**request.param) as driver:
        driver.verify_connectivity()
    return request.param


@fixture(scope="session", params=["Neo4jInterface"])
def db_interface(request, driver_kwds):
    """Fixture to return the experiment interface for the database"""
    match request.param:
        case "Neo4jInterface":
            interface = Neo4jInterface(**driver_kwds)
        case _:
            interface = EmptyInterface()
    yield interface
    interface.close()


@fixture(scope="session", params=["Bacteria", "JackJumper"])
def genotype(request):
    """Fixture to return the phenotype generator from example"""
    return examples.genotypes.__dict__[request.param]


@fixture(scope="session", params=["Random"])
def fitness(request):
    """Fixture to return fitness class from example"""
    return examples.evaluation.__dict__[request.param]


@fixture(scope="session", params=["BasicUniform", "BasicPonderated"])
def algorithm(request):
    """Fixture to return the algorithm definition from example"""
    return examples.algorimths.__dict__[request.param]


@fixture(scope="session", params=[10, 20])
def population_size(request):
    """Parametrization for the number of phenotypes in the population"""
    return request.param


@fixture(scope="session")
def population_gen(genotype, population_size):
    """Fixture to return the population generator"""
    return lambda: [genotype() for _ in range(population_size)]


@fixture(scope="module")
def population(population_gen):
    """Fixture to return a population instance"""
    return population_gen()



"""Module to test evolution experiments"""
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument

import uuid

from pytest import fixture, mark, raises

import gevopy as ea
from gevopy.algorithms import Algorithm
from gevopy.database import Interface
from gevopy.fitness import FitnessModel


# Module fixtures ---------------------------------------------------
@fixture(scope="class", params=[5])
def max_generation(request):
    """Parametrization for the maximum generations to run on experiment"""
    return request.param


@fixture(scope="class", params=[0.8])
def max_score(request):
    """Parametrization for the maximum score to achieve on experiment"""
    return request.param


@fixture(scope="class")
def experiment_name():
    """Fixture to generate an experiment name for testing"""
    return f"Experiment_{uuid.uuid4()}"


@fixture(scope="class")
def session(experiment, population):
    """Fixture to open and return an experiment session for evolution"""
    with experiment.session() as session:
        session.add_phenotypes(population)
        yield session
        session.del_experiment()


@fixture(scope="class")
def execution(session, max_generation, max_score):
    """Fixture to run an experiment and return execution"""
    return session.run(max_generation=max_generation, max_score=max_score)


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Experiment instances attributes"""

    def test_attr_name(self, experiment):
        """Test experiment has a correct 'name' attribute"""
        assert hasattr(experiment, "name")
        assert isinstance(experiment.name, str)

    def test_attr_fitness(self, experiment):
        """Test fitness is instance of FitnessModel"""
        assert hasattr(experiment, "fitness")
        assert isinstance(experiment.fitness, FitnessModel)

    def test_attr_algorithm(self, experiment):
        """Test algorithm is instance of Algorithm"""
        assert hasattr(experiment, "algorithm")
        assert isinstance(experiment.algorithm, Algorithm)

    @mark.parametrize("db_interface", ["Neo4jInterface"], indirect=True)
    def test_neo4j_database(self, experiment):
        """Test experiment has a correct 'database' attribute"""
        assert hasattr(experiment, "database")
        assert isinstance(experiment.database, Interface)

    @mark.parametrize("db_interface", ["EmptyInterface"], indirect=True)
    def test_none_database(self, experiment):
        """Test experiment has a correct 'database' attribute"""
        assert hasattr(experiment, "database")
        assert isinstance(experiment.database, Interface)


class ExecRequirements:
    """Tests group for experiment execution"""

    @mark.parametrize("max_score", [None], indirect=True)
    def test_generation_stop(self, execution, max_generation):
        """Test execution stops when maximum generation is reached"""
        assert execution.generation == max_generation
        assert execution.best_score > 0.0

    @mark.parametrize("max_generation", [None], indirect=True)
    def test_score_stop(self, execution, max_score):
        """Test execution stops when maximum score is reached"""
        assert execution.best_score >= max_score
        assert execution.generation > 0

    def test_attr_generation(self, execution):
        """Test generation attr returns int after first execution"""
        assert hasattr(execution, "generation")
        assert isinstance(execution.generation, int)

    def test_attr_score(self, execution):
        """Test score attr returns float/int after first execution"""
        assert hasattr(execution, "best_score")
        assert isinstance(execution.best_score, (float, int))


class ErrRequirements:
    """Tests group for experiment exceptions requirements"""

    def test_unknown_args(self, session, max_score):
        """Test score attr returns float/int after first execution"""
        with raises(ValueError):
            session.run(unknown_kwarg="something", max_score=max_score)


# Parametrization ---------------------------------------------------
class TestExperiments(AttrRequirements, ExecRequirements, ErrRequirements):
    """Parametrization for genetic evolution 'Experiment'"""

    @fixture(scope="class")
    def experiment(self, experiment_name, db_interface, fitness, algorithm):
        """Parametrization to define the algorithm configuration to use"""
        return ea.Experiment(
            name=experiment_name,
            database=db_interface,
            fitness=fitness(),
            algorithm=algorithm(),
        )



"""Module to test evolution algorithm requirements"""
# pylint: disable=redefined-outer-name

from pytest import fixture, mark

from gevopy import algorithms, genetics
from gevopy.tools import crossover, mutation, selection


# Module fixtures ---------------------------------------------------
@fixture(scope="class", autouse=True)
def score_phenotypes(population):
    """Fixture to add a score to a population of phenotypes"""
    for index, phenotype in enumerate(population):
        phenotype.score = index


@fixture(scope="class")
def result(algorithm, population):
    """Fixture to run an algorithm cycle and return result"""
    return algorithm(population)


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Algorithm instances attributes"""

    def test_attr_selection1(self, algorithm):
        """Test selection1 is instance of tools Selection"""
        assert hasattr(algorithm, "selection1")
        assert isinstance(algorithm.selection1, selection.Selection)

    def test_attr_selection2(self, algorithm):
        """Test selection2 is instance of tools Selection"""
        assert hasattr(algorithm, "selection2")
        assert isinstance(algorithm.selection2, selection.Selection)

    def test_attr_crossover(self, algorithm):
        """Test crossover is instance of tools Crossover"""
        assert hasattr(algorithm, "crossover")
        assert isinstance(algorithm.crossover, crossover.Crossover)

    def test_attr_mutation(self, algorithm):
        """Test mutation is instance of tools Mutation"""
        assert hasattr(algorithm, "mutation")
        assert isinstance(algorithm.mutation, mutation.Mutation)


class ExecutionRequirements:
    """Tests group for Algorithm execution features"""

    def test_return_run(self, result):
        """Test algorithm returns a list of genotype models"""
        assert all(isinstance(x, genetics.GenotypeModel) for x in result)

    @mark.parametrize("survival_rate", [0.0], indirect=True)
    def test_no_ids_keep(self, result, population):
        """Test algorithm return phenotypes are completelly new"""
        assert not any(x in population for x in result)

    @mark.parametrize("survival_rate", [0.5], indirect=True)
    def test_some_ids_change(self, result, population):
        """Test algorithm return phenotypes has new members"""
        assert not all(x in population for x in result)

    @mark.parametrize("survival_rate", [0.5], indirect=True)
    def test_some_ids_keep(self, result, population):
        """Test algorithm return phenotypes are completelly new"""
        assert any(x in population for x in result)

    @mark.parametrize("survival_rate", [1.0], indirect=True)
    def test_all_ids_keep(self, result, population):
        """Test algorithm returns no new phenotypes"""
        assert all(x in population for x in result)


# Parametrization ---------------------------------------------------
class StdParameters:
    """Parametrization for 'Standard' algorithm"""

    @fixture(scope="class", params=["Ponderated", "Uniform"])
    def selection1(self, request):
        """Parametrization to assign the algorithm configuration to use"""
        return selection.__dict__[request.param]()

    @fixture(scope="class", params=["Uniform"])
    def selection2(self, request):
        """Parametrization to assign the algorithm configuration to use"""
        return selection.__dict__[request.param]()

    @fixture(scope="class", params=["TwoPoint"])
    def crossover(self, request):
        """Parametrization to assign the crossover operation to use"""
        return crossover.__dict__[request.param]()

    @fixture(scope="class", params=["SinglePoint"])
    def mutation(self, request):
        """Parametrization to assign the mutation operation to use"""
        return mutation.__dict__[request.param]()

    @fixture(scope="class", params=[0.2, 0.6])
    def survival_rate(self, request):
        """Parametrization to assign the number of phenotypes to survive"""
        return request.param


class TestStandard(StdParameters, AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Standard' algorithm"""

    @fixture(scope="class")
    def selections(self, selection1, selection2):
        """Parametrization to define the selection configuration to use"""
        return dict(selection1=selection1, selection2=selection2)

    @fixture(scope="class")
    def evolution(self, crossover, mutation):
        """Parametrization to define the evolution configuration to use"""
        return dict(crossover=crossover, mutation=mutation)

    @fixture(scope="class")
    def extra(self, survival_rate):
        """Parametrization to define the extra configuration to use"""
        return dict(survival_rate=survival_rate)

    @fixture(scope="class")
    def algorithm(self, selections, evolution, extra):
        """Parametrization to define the algorithm configuration to use"""
        return algorithms.Standard(**selections, **evolution, **extra)



"""Examples module for users on how to create algorithms and testing"""
from gevopy import algorithms
from gevopy.tools import crossover, mutation, selection


# ------------------------------------------------------------------
# Basic Uniform Selection ------------------------------------------
# This algorithm selects both matching partners using an uniform
# distribution. Due to the fact that fitness does not play any role
# on the selection of phenotypes, adaptation to the environment is
# fully random.
class BasicUniform(algorithms.Standard):
    selection1 = selection.Uniform()
    selection2 = selection.Uniform()
    crossover = crossover.OnePoint()
    mutation = mutation.SinglePoint()


# ------------------------------------------------------------------
# Simple Ponderated Selection --------------------------------------
# This algorithm selects one matching partners using a ponderated
# distribution and the second with an uniform. Ponderated gives all
# phenotypes a chance to reproduce proportional to the fitness score.
# However, the posibility that best matches best is reduced agains
# two times ponderated.
class BasicPonderated(algorithms.Standard):
    selection1 = selection.Ponderated()
    selection2 = selection.Uniform()
    crossover = crossover.OnePoint()
    mutation = mutation.SinglePoint()



"""Module to test crossover utilities for evolution algorithms."""
# pylint: disable=redefined-outer-name

import numpy as np
from pytest import fixture, mark

from gevopy import utils


@fixture(scope="function", autouse=True)
def apply_crossover(chromosome_1, chromosome_2, points):
    """Crosses chromosomes on the indicated list of points"""
    utils.cross_chromosomes(chromosome_1, chromosome_2, points)


@mark.parametrize("chromosome_1", [np.array([11, 12, 13, 14])])
@mark.parametrize("chromosome_2", [np.array([21, 22, 23, 24])])
@mark.parametrize("points", [[0, 0, 0]])
def test_000_cross_chromosomes(chromosome_1, chromosome_2):
    """Tests arrays are crossed at positions 000"""
    assert all(chromosome_1 == [21, 22, 23, 24])
    assert all(chromosome_2 == [11, 12, 13, 14])


@mark.parametrize("chromosome_1", [np.array([11, 12, 13, 14])])
@mark.parametrize("chromosome_2", [np.array([21, 22, 23, 24])])
@mark.parametrize("points", [[0, 0, 1]])
def test_001_cross_chromosomes(chromosome_1, chromosome_2):
    """Tests arrays are crossed at positions 001"""
    assert all(chromosome_1 == [11, 22, 23, 24])
    assert all(chromosome_2 == [21, 12, 13, 14])


@mark.parametrize("chromosome_1", [np.array([11, 12, 13, 14])])
@mark.parametrize("chromosome_2", [np.array([21, 22, 23, 24])])
@mark.parametrize("points", [[0, 0, 4]])
def test_004_cross_chromosomes(chromosome_1, chromosome_2):
    """Tests arrays are crossed at positions 004"""
    assert all(chromosome_1 == [11, 12, 13, 14])
    assert all(chromosome_2 == [21, 22, 23, 24])


@mark.parametrize("chromosome_1", [np.array([11, 12, 13, 14])])
@mark.parametrize("chromosome_2", [np.array([21, 22, 23, 24])])
@mark.parametrize("points", [[0, 1, 2]])
def test_012_cross_chromosomes(chromosome_1, chromosome_2):
    """Tests arrays are crossed at positiosn 012"""
    assert all(chromosome_1 == [21, 12, 23, 24])
    assert all(chromosome_2 == [11, 22, 13, 14])


@mark.parametrize("chromosome_1", [np.array([11, 12, 13, 14])])
@mark.parametrize("chromosome_2", [np.array([21, 22, 23, 24])])
@mark.parametrize("points", [[1, 2, 3]])
def test_123_cross_chromosomes(chromosome_1, chromosome_2):
    """Tests arrays are crossed at positions 123"""
    assert all(chromosome_1 == [11, 22, 13, 24])
    assert all(chromosome_2 == [21, 12, 23, 14])



"""Module to test mutation operations for evolution algorithms."""
# pylint: disable=redefined-outer-name

from inspect import ismethod, signature

from pytest import fixture, mark

from gevopy import genetics
from gevopy.tools import mutation


# Module fixtures ---------------------------------------------------
@fixture(scope="class")
def phenotype(mutation, original):
    """Fixture to return a mutated phenotype from base"""
    return mutation(original)


@fixture(scope="class")
def original(genotype):
    """Fixture to generate a base phenotype to mutate"""
    return genotype()


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Mutation instances attributes"""

    def test_is_callable(self, mutation):
        """Test mutation is a callable with arity 1"""
        assert callable(mutation)
        assert len(signature(mutation).parameters) == 1

    def test_attr_mutate_chromosome(self, mutation):
        """Test attribute mutate_chromosome is a method"""
        assert hasattr(mutation, "mutate_chromosome")
        assert ismethod(mutation.mutate_chromosome)


class ExecutionRequirements:
    """Tests group for Mutation execution features"""

    def test_keeps_genotype(self, original, phenotype):
        """Test mutation does not alter phenotype"""
        assert isinstance(phenotype, genetics.GenotypeModel)
        assert type(original) is type(phenotype)

    def test_mutated_diff(self,  original, phenotype):
        """Test mutation returns different object than phenotype"""
        assert original is not phenotype

    def test_all_ids_keep(self,  original, phenotype):
        """Test mutation does not alter phenotype id"""
        assert original.id == phenotype.id


# Parametrization ---------------------------------------------------
class TestSinglePoint(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'SinglePoint' mutation"""

    @fixture(scope="class", params=[0.1, 0.2, 0.5])
    def mutation(self, request):
        """Parametrization to define the mutation method to use"""
        return mutation.SinglePoint(mutpb=request.param)

    @mark.parametrize("mutation", [0.0], indirect=True)
    def test_mutpb_000(self, original, phenotype):
        """Tests that no mutation does not modity chromosome"""
        assert original.chromosome == phenotype.chromosome

    @mark.parametrize("mutation", [1.0], indirect=True)
    def test_mutpb_100(self, original, phenotype):
        """Tests that mutation does modity chromosome"""
        assert original.chromosome != phenotype.chromosome



"""The :mod:`algorithms` module is intended to contain some specific
algorithms in order to execute the most common evolutionary algorithms.

The methods used here are more for convenience than reference as the
implementation of every evolutionary algorithm may vary infinitely. Most
of the algorithms in this module use operators registered in the tools
package.

A rework form:
https://github.com/DEAP/deap/blob/master/deap/algorithms.py
"""

import functools
import logging
import math
from abc import ABC, abstractmethod

from pydantic import BaseModel

from gevopy import tools
from gevopy.tools.crossover import Crossover
from gevopy.tools.mutation import Mutation
from gevopy.tools.selection import Selection

# https://docs.python.org/3/howto/logging-cookbook.html
module_logger = logging.getLogger(__name__)


class Algorithm(BaseModel, ABC):
    """Base abstract class for algorithms generation and identification.
    An algorithm is composed generally by at least 3 components:
     - Selection: Tool to select phenotypes
     - Crossover: Tool to crossover phenotypes
     - Mutation: Tool to mutate phenotypes

    This components can be assigned to an algorithm class by Inheritance
    using the provided classes `Has<Attribute>`. Instances from an algorithm
    can be created and edited after, but their values are limited to the
    corresponding classes from `tools` subpackage.

    .. note::
        Some optimization problems might not be suitable for crossover or
        mutation operations.

    When an algoritm has the property crossover, you can usually configure
    a second `selection` argument `selection2` to control how the algorithm
    selects the secondary phenotype.
    """

    class Config:
        # pylint: disable=missing-class-docstring
        # pylint: disable=too-few-public-methods
        keep_untouched = (functools.cached_property,)

    @functools.cached_property
    def logger(self):
        """Experiment logger, used to trace and print experiment info.
        :return: Experiment.Logger
        """
        return logging.getLogger(f"{self.__class__}")

    def __call__(self, population):
        """Executes the algorithm to return a population offspring.
        :param population: List of phenotypes
        :returns: A list of varied phenotypes
        """
        return self.run_cycle(tools.Pool(population))

    @abstractmethod
    def run_cycle(self, phenotypes):
        """Executes an algorithm cyle to return a population offspring.
        :param phenotypes: Pool of ordered phenotypes by score
        :returns: A list of varied phenotypes
        """
        raise NotImplementedError


class HasSelection1(Algorithm):
    """Extend class Algorithm with a second selection round of phenotypes."""
    selection1: Selection

    def __init__(self, **data) -> None:
        """Generic constructor for genetic algorithms."""
        super().__init__(**data)

    def select(self, phenotypes, n):
        """Extends algorithm call method with phenotypes selection operation.
        :param phenotypes: Pool of evaluated phenotypes
        :param n: Number of phenotype to select from pool
        :returns: A list of varied phenotypes
        """
        self.logger.debug("selection1:\t%s", self.selection1)
        selected1 = self.selection1(phenotypes, n)
        self.logger.debug("selection1=\t%s", selected1)
        return selected1


class HasSelection2(HasSelection1, Algorithm):
    """Extend class Algorithm with a second selection round of phenotypes."""
    selection2: Selection

    def select(self, phenotypes, n):
        """Extends algorithm call method with phenotypes selection operation.
        :param phenotypes: Pool of evaluated phenotypes
        :param n: Number of phenotype to select from pool
        :returns: A list of varied phenotypes
        """
        selected1 = super().select(phenotypes, n)
        self.logger.debug("selection2:\t%s", self.selection2)
        selected2 = self.selection2(phenotypes, n)
        self.logger.debug("selection2=\t%s", selected2)
        return selected1, selected2


class HasCrossover(HasSelection2, Algorithm):
    """Extend class Algorithm with crossover properties and methods."""
    crossover: Crossover

    def cross(self, selected1, selected2):
        """Extends algorithm call method with phenotypes crossover operation.
        :param selected1: List 1 of phenotypes to cross
        :param selected2: List 2 of phenotypes to cross
        :returns: A list of varied phenotypes
        """
        self.logger.debug("crossover:\t%s", self.crossover)
        selections = zip(selected1, selected2)
        offspring = [self.crossover(x, y)[0] for x, y in selections]
        self.logger.debug("offspring=\t%s", offspring)
        return offspring


class HasMutation(HasSelection1, Algorithm):
    """Extend class Algorithm with mutation properties and methods."""
    mutation: Mutation

    def mutate(self, phenotypes):
        """Extends algorithm call method with phenotypes mutation operation.
        :param phenotypes: List of evaluated phenotypes
        :returns: A list of varied phenotypes
        """
        self.logger.debug("mutation:\t%s", self.mutation)
        offspring = [self.mutation(phenotype) for phenotype in phenotypes]
        self.logger.debug("offspring=\t%s", offspring)
        return offspring


class HasSurvivalRate(HasSelection1, Algorithm):
    """Extend class Algorithm with survival properties and methods. When
    subclassing this methods, the algorithm protects the most performant
    phenotypes from the previous generation.
    """
    survival_rate: float = 0.4

    def skim(self, phenotypes):
        """Extends algorithm call method to protect a ratio of phenotypes.
        :param phenotypes: List of evaluated phenotypes
        :returns: A tuple with survivors and discarted phenotypes
        """
        self.logger.debug("survival_rate:\t%s", self.survival_rate)
        n_survivors = math.ceil(len(phenotypes) * self.survival_rate)
        survivors = phenotypes[:n_survivors]
        self.logger.debug("survivors=\t%s", survivors)
        return survivors, phenotypes[n_survivors:]


class Standard(HasSurvivalRate, HasCrossover, HasMutation, Algorithm):
    """This algorithm reproduce the standard evolutionary algorithm.

    First, a selection procedure is applied to replace a fraction of the
    population with deep copies of the parental population. Second, applies
    crossover between selected copies, generating new phenotypes. Third,
    applies mutation to crossover offspring, generating new phenotypes.

    Before the selection procedure, the percentage of phenotypes with lowest
    score defined by (1.0 - `survival_rate`) are removed. Phenotypes filtered
    by the survival process are replaced by a selection of survivors.
    """
    # pylint: disable=too-many-arguments

    def run_cycle(self, phenotypes):
        """Execution entry point for the algorithm.
        :param phenotypes: List of evaluated phenotypes
        :returns: A list of varied phenotypes
        """
        self.logger.debug("survival_phenotypes=\t%s", phenotypes)
        survivors, rest = self.skim(phenotypes)
        selected1, selected2 = self.select(phenotypes, len(rest))
        offspring = self.cross(selected1, selected2)
        offspring = self.mutate(offspring)
        self.logger.debug("simple_offspring=\t%s", offspring)
        return survivors + offspring



"""Evolution Algorithms, Tools subpackage."""

import collections
import heapq

import sortedcontainers


class Pool(sortedcontainers.SortedKeyList):
    """Pool is a sorted list of dictionaries where items are composed
    by a *score* and a *phenotype.
    """

    def __init__(self, iterable=None):
        """Constructor for a pool as sorted mutable sequence.
        :param iterable: Iterable of values to initialize the sorted-key list
        """
        super().__init__(iterable, key=lambda item: -item.score)

    def scores(self):
        """Returns a sorted list of phenotype scores from highest to lowest.
        :return: List of numbers (int or float)
        """
        return [item.score for item in self]

    def append(self, value):
        """Method append is unsupported on chromosome types"""
        raise TypeError(f"Unsupported operation by '{self.__class__}'")

    def extend(self, values):
        """Method extend is unsupported on chromosome types"""
        raise TypeError(f"Unsupported operation by '{self.__class__}'")

    def insert(self, index, value):
        """Method insert is unsupported on chromosome types"""
        raise TypeError(f"Unsupported operation by '{self.__class__}'")

    def reverse(self):
        """Method reverse is unsupported on chromosome types"""
        raise TypeError(f"Unsupported operation by '{self.__class__}'")


class HallOfFame(collections.abc.Sequence):
    """The hall of fame contains the best individual that ever lived in an
    evolution context (for example an algorithm). It is sorted at all time
    so that the first element of the hall of fame is the item that contains
    the phenotype with the best score.

    The update is made so that old phenotypes have priority on new phenotypes.
    The class :class:`HallOfFame` provides an interface similar to a list
    (without being one completely). It is possible to retrieve its length,
    to iterate on it forward and backward and to get an item or a slice.
    """

    def __init__(self, maxsize):
        """Hall of fame constructor.
        :param maxsize: Maximum number of phenotypes to keep in the hall
        """
        self.__items = []
        self.maxsize = maxsize

    @property
    def maxsize(self):
        """Property that defines the maximum ammount of phenotypes in the hall.
        :return: Maximum allowed number of phenotypes in the hall
        """
        return self.__maxsize

    @maxsize.setter
    def maxsize(self, value):
        match value:
            case _ if not isinstance(value, int):
                raise ValueError("Expected type 'int' for maxsize")
            case _ if not value > 0:
                raise ValueError("Expected maxsize value higher than 0")
            case _ if value < len(self):
                self.__items = self[:value]
        self.__maxsize = value

    def __iter__(self):
        return iter(self.__items)

    def __contains__(self, value):
        return value in self.__items

    def __getitem__(self, index):
        return self.__items.__getitem__(index)

    def __len__(self):
        return self.__items.__len__()

    def __repr__(self) -> str:
        return self.__items.__repr__()

    @classmethod
    def __get_validators__(cls):
        yield cls.class_validator

    @classmethod
    def class_validator(cls, value):
        """Validates the value is a correct HallOfFame type."""
        if not isinstance(value, cls):
            raise TypeError("'HallOfFame' type required")
        return value

    def update(self, pool):
        """Updates phenotypes to the hall of fame by replacing the worst
        phenotypes in the hall by the best phenotypes present in pool. Old
        phenotypes have preference over new phenotypes.
        :param pool: A pool of phenotypes
        """
        if not isinstance(pool, Pool):
            raise TypeError("Expected type 'Pool' for phenotypes pool")
        records = heapq.merge(self, pool, key=lambda x: -x.score)
        self.__items = [x for _, x in zip(range(self.maxsize), records)]



"""Examples module for users on how to generate genotypes and testing

Note that the example names are not really matching with the reality as it
might be impossible to fully represent biological genotypes. However they
might serve well for ilustration purposes.
"""
from gevopy import random
from gevopy.genetics import Field, GenotypeModel, Haploid, Diploid


# ------------------------------------------------------------------
# Genotype ---------------------------------------------------------
# This genotypes are known to have asexual reproduction and with a
# unique chromosome. In this example values can be only between 0 and 1
# See https://pydantic-docs.helpmanual.io/usage/models/
class Bacteria(GenotypeModel):
    """Simple and most basic haploid genotype"""
    chromosome: Haploid = Field(default_factory=lambda: random.haploid(12))


# ------------------------------------------------------------------
# Genotype ---------------------------------------------------------
# Although males are know to be composed by haploids, this is a good
# example of how to design a chromosome where values can take more
# values than 1 or 0 (0-3) due to that each value in the chromosome
# might be composed by 2 bits ([00][00]-...-[00][00]).
# See https://pydantic-docs.helpmanual.io/usage/models/
class JackJumper(GenotypeModel):
    """This anst are known to have a single pair of chromosomes"""
    chromosome: Diploid = Field(default_factory=lambda: random.diploid(12))



"""Evolution algorithm module to generate random units, for example random
chromosomes.
"""

import numpy as np

from gevopy import genetics


def haploid(size):
    """Returns a random standard Haploid chromosome.
    :param size: Integer with chromosome size
    """
    data = np.random.randint(2 ** 1, size=size, dtype="uint8")
    return genetics.Haploid(data)


def diploid(size):
    """Returns a random standard Diploid chromosome.
    :param size: Integer with chromosome size
    """
    data = np.random.randint(2 ** 2, size=size, dtype="uint8")
    return genetics.Diploid(data)


def triploid(size):
    """Returns a random standard Triploid chromosome.
    :param size: Integer with chromosome size
    """
    data = np.random.randint(2 ** 3, size=size, dtype="uint8")
    return genetics.Triploid(data)



"""Module to test selection operations for evolution algorithms."""
# pylint: disable=redefined-outer-name

from inspect import signature

from pytest import fixture, mark

from gevopy.tools import Pool, selection


# Module fixtures ---------------------------------------------------
@fixture(scope="class", autouse=True)
def score_phenotypes(population):
    """Fixture to add a score to a population of phenotypes"""
    for index, phenotype in enumerate(population):
        phenotype.score = index


@fixture(scope="class")
def pool(population):
    """Fixture to return a pool of phenotypes"""
    return Pool(population)


@fixture(scope="class", params=[0, 1, 5, 20])
def selection_size(request):
    """Parametrization for the number of phenotypes to select"""
    return request.param


@fixture(scope="class")
def selected(selection, pool, selection_size):
    """Fixture to return a selection of phenotypes"""
    return selection(pool, selection_size)


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Selection instances attributes"""

    def test_is_callable(self, selection):
        """Test selection is a callable with arity 2"""
        assert callable(selection)
        assert len(signature(selection).parameters) == 2


class ExecutionRequirements:
    """Tests group for Selection execution features"""

    def test_returned_size(self, selected, selection_size):
        """Test returned list length matches the selection size"""
        assert len(selected) == selection_size

    def test_selected_from_pool(self, selected, pool):
        """Test all returned objects come from the original pool"""
        assert all(x in pool for x in selected)


# Parametrization ---------------------------------------------------
class TestPonderated(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Ponderated' selection"""

    @fixture(scope="class")
    def selection(self):
        """Parametrization to define the selection method to use"""
        return selection.Ponderated()

    @mark.skip(reason="TODO: Help needed for asserts")
    def test_selected(self, selected, pool):  # pylint: disable=unused-argument
        """Tests the correct selection of phenotypes"""
        assert NotImplementedError


class TestUniform(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Uniform' selection"""

    @fixture(scope="class")
    def selection(self):
        """Parametrization to define the selection method to use"""
        return selection.Uniform()

    @mark.skip(reason="TODO: Help needed for asserts")
    def test_selected(self, selected, pool):  # pylint: disable=unused-argument
        """Tests the correct selection of phenotypes"""
        assert NotImplementedError


class TestBest(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Best' selection"""

    @fixture(scope="class")
    def selection(self):
        """Parametrization to define the selection method to use"""
        return selection.Best()

    def test_selected(self, selected, pool):
        """Tests the correct selection of phenotypes"""
        assert all(ph == pool[0] for ph in selected)


class TestWorst(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Worst' selection"""

    @fixture(scope="class")
    def selection(self):
        """Parametrization to define the selection method to use"""
        return selection.Worst()

    def test_selected(self, selected, pool):
        """Tests the correct selection of phenotypes"""
        assert all(ph == pool[-1] for ph in selected)


class TestTournaments(AttrRequirements, ExecutionRequirements):
    """Parametrization for 'Torunaments' selection"""

    @fixture(scope="class")
    def selection(self, tournsize):
        """Parametrization to define the selection method to use"""
        return selection.Tournaments(tournsize)

    @fixture(scope="class", params=[2, 5])
    def tournsize(self, request):
        """Parametrization to define the tournament size"""
        return request.param

    @mark.skip(reason="TODO: Help needed for asserts")
    def test_selected(self, selected, pool):  # pylint: disable=unused-argument
        """Tests the correct selection of phenotypes"""
        assert NotImplementedError



"""Module to test HallOfFamme utility for evolution algorithms."""
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument

from pytest import fixture, mark

from examples.genotypes import Bacteria as Genotype
from gevopy import tools


@fixture(scope="function")
def halloffame(maxsize):
    """Instantiates and returns a HallOfFame"""
    return tools.HallOfFame(maxsize)


@fixture(scope="function", autouse=True)
def update(halloffame, population1, population2):
    """Updates Hall Of Fame with pools of phenotypes"""
    halloffame.update(tools.Pool(population1))
    halloffame.update(tools.Pool(population2))


@fixture(scope="function")
def newsize(request, halloffame, update):
    """Updates Hall Of Fame with pool of phenotypes"""
    halloffame.maxsize = request.param
    return request.param


@mark.parametrize("maxsize", [1, 3, 5, 9])
@mark.parametrize("population1", [[Genotype(score=x) for x in [0, 1, 5, 7]]])
@mark.parametrize("population2", [[Genotype(score=x) for x in [0, 1, 5, 7]]])
def test_maxsize(halloffame, maxsize):
    """Test the length of the Hall Of Fame never passes maxsize"""
    assert len(halloffame) <= maxsize


@mark.parametrize("maxsize", [1, 3, 5, 9])
@mark.parametrize("population1", [[Genotype(score=x) for x in [0, 1, 5, 7]]])
@mark.parametrize("population2", [[Genotype(score=x) for x in [0, 1, 5, 7]]])
def test_order(halloffame):
    """Test the elements in the Hall Of Fame are ordered by score"""
    sorted_scores = sorted([x.score for x in halloffame], reverse=True)
    assert [x.score for x in halloffame] == sorted_scores


@mark.parametrize("newsize", [1, 9], indirect=True)
@mark.parametrize("maxsize", [1, 5, 9])
@mark.parametrize("population1", [[Genotype(score=x) for x in [0, 1, 5, 7]]])
@mark.parametrize("population2", [[Genotype(score=x) for x in [0, 1, 5, 7]]])
def test_newsize(halloffame, newsize):
    """Test the length of the Hall Of Fame never passes newsize"""
    assert len(halloffame) <= newsize



"""This modules interfaces the application with neo4j sotred data."""
import abc
import contextlib
import functools
import logging

import neo4j

import gevopy.config

# Database logger and configuration
module_logger = logging.getLogger(__name__)
config = gevopy.config.settings['database']


# Database interfaces -----------------------------------------------

class Interface(abc.ABC):
    """Abstract class for Database Interface Object."""

    @abc.abstractmethod
    def close(self,  *args, **kwds):
        """Function to close the database driver.
        :param args: Same as database Driver close possitional arguments
        :param kwds: Same as database Driver close key arguments
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def closed(cls, session):
        """Function to evaluate if session from interface is closed or open.
        :param session: Database Session to evaluate
        :return: True if closed, otherwise False
        """
        raise NotImplementedError

    @classmethod
    def __get_validators__(cls):
        yield cls.class_validator

    @classmethod
    def class_validator(cls, value):
        """Validates the value is a correct Interface type."""
        if not isinstance(value, cls):
            raise TypeError("'Interface' type required")
        return value

    # abc.abstractcontextmanagermethod
    def session(self, *args, **kwds):
        """Function to generate a context session to interface the experiment.
        :param args: Same as database Driver session possitional arguments
        :param kwds: Same as database Driver session key arguments
        :return: Database session container with interface methods
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def add_phenotypes(cls, container, phenotypes):
        """Creates new phenotypes in the database and returns the ids.
        :param container: Session container for database session
        :param phenotypes: List/iter of serialized phenotypes to store
        :return: List of ids from the created phenotypes
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_phenotypes(cls, container, ids):
        """Reads the database and returns the matching id phenotypes.
        :param container: Session container for database session
        :param ids: Ids from the phenotypes to collect
        :return: Serialized phenotypes matching the input ids
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def del_phenotypes(cls, container, ids):
        """Deletes from database and returns the matching id phenotypes.
        :param container: Session container for database session
        :param ids: Ids from the phenotypes to collect
        :return: Deleted phenotypes matching the input ids
        """
        raise NotImplementedError


class Neo4jInterface(Interface):
    """Neo4j database interface as evolution graph. It replacess all neo4j
    session methods by customised transactions ready to execute.

    Similar to the neo4j session, this class has context manager properties,
    therefore its methods must be executed inside a 'with' context.
    """
    logger = logging.getLogger(f"{__name__}.Neo4jInterface")

    def __init__(self, *args, **kwds):
        """Function to generate the neo4j interface object.
        :param args: Same as neo4j.Driver open possitional arguments
        :param kwds: Same as neo4j.Driver open key arguments
        """
        self.logger.debug('Opening driver: %s %s', args, kwds)
        self.__driver = neo4j.GraphDatabase.driver(*args, **kwds)

    def close(self,  *args, **kwds):
        """Function to close the neo4j driver.
        :param args: Same as neo4j.Driver close possitional arguments
        :param kwds: Same as neo4j.Driver close key arguments
        """
        self.logger.debug('Closing driver: %s %s', args, kwds)
        self.__driver.close(*args, **kwds)

    @classmethod
    def closed(cls, session):
        """Function to evaluate if session from interface is closed or open.
        :param session: neo4j Session to evaluate
        :return: True if closed, otherwise False
        """
        return session.closed()

    @contextlib.contextmanager
    def session(self, *args, **kwds):
        """Function to generate a context session to interface the experiment.
        :param args: Same as neo4j.Driver session possitional arguments
        :param kwds: Same as neo4j.Driver session key arguments
        :return: Database session container with interface methods
        """
        self.logger.debug('Session call: %s %s', args, kwds)
        with self.__driver.session(*args, **kwds) as db_session:
            self.logger.debug("Open neo4j database session: %s", db_session)
            yield SessionContainer(db_session, interface=self)
            self.logger.debug("Exit neo4j database session: %s", db_session)

    @classmethod
    def add_phenotypes(cls, container, phenotypes):
        """Creates new phenotypes in the database and returns the ids.
        :param container: Session container for neo4j session
        :param phenotypes: List/iter of serialized phenotypes to store
        :return: List of ids from the created phenotypes
        """
        phenotypes = list(phenotypes)
        cls.logger.debug('Adding phenotypes %s', phenotypes)
        return container.session.execute_write(add_phenotypes, phenotypes)

    @classmethod
    def get_phenotypes(cls, container, ids):
        """Reads the database and returns the matching id phenotypes.
        :param container: Session container for neo4j session
        :param ids: Ids from the phenotypes to collect
        :return: Serialized phenotypes matching the input ids
        """
        ids = list(str(id) for id in ids)
        cls.logger.debug('Getting phenotypes %s', ids)
        return container.session.execute_read(get_phenotypes, ids)

    @classmethod
    def del_phenotypes(cls, container, ids):
        """Deletes from database and returns the matching id phenotypes.
        :param container: Session container for database session
        :param ids: Ids from the phenotypes to delete
        :return: Deleted phenotypes matching the input ids
        """
        ids = list(str(id) for id in ids)
        cls.logger.debug('Deleting phenotypes %s', ids)
        return container.session.execute_write(del_phenotypes, ids)

    @classmethod
    def del_experiment(cls, container, name):
        """Transaction to delete experiment and related data.
        :param container: Session container for database session
        :param name: Id of the experiment to delete
        """
        cls.logger.debug('Deleting experiment %s', name)
        return container.session.execute_write(del_experiment, name)


class EmptyInterface(Interface):
    """Neo4j database interface as evolution graph. It replacess all neo4j
    session methods by customised transactions ready to execute.

    Similar to the neo4j session, this class has context manager properties,
    therefore its methods must be executed inside a 'with' context.
    """
    logger = logging.getLogger(f"{__name__}.EmptyInterface")

    def __init__(self):
        """Function to generate the empty mock interface object.
        """
        self.logger.debug('Opening empty interface driver')

    def close(self,  *args, **kwds):
        """Function to close the empty mock interface object.
        """
        self.logger.debug('Closing empty interface driver')

    @classmethod
    def closed(cls, session):
        """Function to evaluate if session from interface is closed or open.
        :param session: Mock session to evaluate
        :return: True if closed, otherwise False
        """
        return ~session['open']

    @contextlib.contextmanager
    def session(self, *args, **kwds):
        """Function to generate a context session to interface the experiment.
        :param args: Ignored
        :param kwds: Ignored
        :return: Database session container with interface methods
        """
        self.logger.debug('Session call: %s %s', args, kwds)
        db_session = dict(open=True)
        self.logger.debug("Open mock database session: %s", db_session)
        yield SessionContainer(db_session, interface=self)
        db_session['open'] = False
        self.logger.debug("Exit mock database session: %s", db_session)

    @classmethod
    def add_phenotypes(cls, _container, phenotypes):
        """Creates new phenotypes in the database and returns the ids.
        :param container: Session container (Not used)
        :param phenotypes: List/iter of serialized phenotypes to store
        :return: List of ids from the created phenotypes
        """
        phenotypes = list(phenotypes)
        cls.logger.debug('Adding phenotypes %s', phenotypes)
        return list(str(p.id) for p in phenotypes)

    @classmethod
    def get_phenotypes(cls, _container, ids):
        """Reads the database and returns the matching id phenotypes.
        :param container: Session container (Not used)
        :param ids: Ids from the phenotypes to collect
        :return: Serialized phenotypes matching the input ids
        """
        ids = list(str(id) for id in ids)
        cls.logger.debug('Getting phenotypes %s', ids)
        return []

    @classmethod
    def del_phenotypes(cls, _container, ids):
        """Deletes from database and returns the matching id phenotypes.
        :param container: Session container (Not used)
        :param ids: Ids from the phenotypes to delete
        :return: Deleted phenotypes matching the input ids
        """
        ids = list(str(id) for id in ids)
        cls.logger.debug('Deleting phenotypes %s', ids)
        return []

    @classmethod
    def del_experiment(cls, _container, name):
        """Transaction to delete experiment and related data.
        :param container: Session container (Not used)
        :param name: Id of the experiment to delete
        """
        cls.logger.debug('Deleting experiment %s', name)


# Session Containers ------------------------------------------------

class AbstractSession(abc.ABC):
    """Abstract class for Database Interface Session Object."""

    def __init__(self, session, interface):
        """Generic constructor for Database Session Container Objects.
        :param session: Real session from the database
        :param interface: Database interface with evolution methods
        """
        self.interface = interface
        self.session = session

    @classmethod
    def require_session(cls, meth):
        """Decorator to ensure the method/function is called in context."""
        @functools.wraps(meth)
        def wrapped_method(self, *args, **kwds):
            if not self.in_session:
                raise RuntimeError(f"{meth} out of {self.__class__} context")
            return meth(self, *args, **kwds)
        return wrapped_method

    @property
    @abc.abstractmethod
    def in_session(self):
        """Returns if the execution pointer is inside the instance context.
        :return: True if the call is inside instance context
        """
        raise NotImplementedError

    @abc.abstractmethod
    def add_phenotypes(self, phenotypes):
        """Creates new phenotypes in the database and returns the ids.
        :param phenotypes: List/iter of serialized phenotypes to store
        :return: List of ids from the created phenotypes
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_phenotypes(self, ids):
        """Reads the database and returns the matching id phenotypes.
        :param ids: Ids from the phenotypes to collect
        :return: Serialized phenotypes matching the input ids
        """
        raise NotImplementedError

    @abc.abstractmethod
    def del_phenotypes(self, ids):
        """Deletes from database and returns the matching id phenotypes.
        :param ids: Ids from the phenotypes to delete
        :return: Deleted phenotypes matching the input ids
        """
        raise NotImplementedError

    @abc.abstractmethod
    def del_experiment(self, name):
        """Transaction to delete experiment and related data.
        :param name: Id of the experiment to delete
        """
        raise NotImplementedError


class SessionContainer(AbstractSession):
    """Generic Database Session Container for Database Interface"""

    @property
    def in_session(self):
        """Returns if the execution pointer is inside the instance context.
        :return: True if the call is inside instance context
        """
        return ~self.interface.closed(self.session)

    @AbstractSession.require_session
    def add_phenotypes(self, phenotypes):
        """Creates new phenotypes in the database and returns the ids.
        :param phenotypes: List/iter of serialized phenotypes to store
        :return: List of ids from the created phenotypes
        """
        return self.interface.add_phenotypes(self, phenotypes)

    @AbstractSession.require_session
    def get_phenotypes(self, ids):
        """Reads the database and returns the matching id phenotypes.
        :param ids: Ids from the phenotypes to collect
        :return: Serialized phenotypes matching the input ids
        """
        return self.interface.get_phenotypes(self, ids)

    @AbstractSession.require_session
    def del_phenotypes(self, ids):
        """Deletes from database and returns the matching id phenotypes.
        :param ids: Ids from the phenotypes to delete
        :return: Deleted phenotypes matching the input ids
        """
        return self.interface.del_phenotypes(self, ids)

    @AbstractSession.require_session
    def del_experiment(self, name):
        """Transaction to delete experiment and related data.
        :param name: Id of the experiment to delete
        """
        return self.interface.del_experiment(self, name)


# NEO4J Transactions ------------------------------------------------

@neo4j.unit_of_work(timeout=config['timeout'])
def add_phenotypes(tx, phenotypes):
    """Transaction function to create new phenotypes in the database.
    :param tx: Neo4j transaction object
    :param phenotypes: List of serialized phenotype to store
    :return: List of ids from the created phenotypes
    """
    query = (
        "UNWIND $phenotypes AS phenotype "
        "MERGE (x:Phenotype { id: phenotype.id }) "
        "SET x = phenotype, x.parents = null, x.experiment = null "
        "WITH phenotype, x as child "
        "CALL { "
        "  WITH child, phenotype "
        "  UNWIND phenotype.parents as parent_id "
        "  MATCH (parent:Phenotype { id: parent_id }) "
        "  MERGE (parent)-[:HAS_CHILD]->(child) "
        "} "
        "WITH child as x, phenotype.experiment as experiment "
        "WHERE NOT experiment IS NULL "
        "CALL { "
        "  WITH x, experiment "
        "  MERGE (e:Experiment { name: experiment }) "
        "  MERGE (x)-[:IN_EXPERIMENT]->(e) "
        "} "
    )
    tx.run(query, phenotypes=phenotypes)
    return [p['id'] for p in phenotypes]


@neo4j.unit_of_work(timeout=config['timeout'])
def get_phenotypes(tx, ids):
    """Transaction to returns the matching id phenotypes.
    :param tx: Neo4j transaction object
    :param ids: Ids from the phenotypes to collect
    :return: Serialized phenotypes matching the input ids
    """
    query = (
        "MATCH (x:Phenotype) WHERE x.id IN $phenotypes_ids "
        "OPTIONAL MATCH (x)-[:IN_EXPERIMENT]->(e:Experiment) "
        "OPTIONAL MATCH (y)-[:HAS_CHILD]->(x) "
        "WITH x, e, collect(y.id) as p "
        "RETURN x{.*, .score, experiment:e.name, parents:p } "
    )
    result = tx.run(query, phenotypes_ids=ids)
    return [dict(record["x"]) for record in result]


@neo4j.unit_of_work(timeout=config['timeout'])
def del_phenotypes(tx, ids):
    """Transaction to delete the matching id phenotypes.
    :param tx: Neo4j transaction object
    :param ids: Ids from the phenotypes to collect
    :return: Ids of the matched and deleted phenotypes
    """
    query = (
        "MATCH (x:Phenotype) WHERE x.id IN $phenotypes_ids "
        "WITH x as phenotype, x.id AS id "
        "DETACH DELETE phenotype "
        "RETURN id "
    )
    result = tx.run(query, phenotypes_ids=ids)
    return [record["id"] for record in result]


@neo4j.unit_of_work(timeout=config['timeout'])
def del_experiment(tx, name):
    """Transaction to delete experiment and related data.
    :param tx: Neo4j transaction object
    :param name: Id of the experiment to delete
    """
    query = (
        "MATCH (e:Experiment { name: $experiment_name }) "
        "OPTIONAL MATCH (x)-[:IN_EXPERIMENT]->(e) "
        "DETACH DELETE e, x "
    )
    tx.run(query, experiment_name=name)



"""Module to test fitness requirements for evolution algorithms"""
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument

from inspect import ismethod, signature

from pytest import fixture, mark


# Module fixtures ---------------------------------------------------
@fixture(scope="class", params=[False, True])
def use_cache(request):
    """Fixture to enable or disable cache on fitness evaluation"""
    return request.param


@fixture(scope="class", params=['synchronous', 'threads', 'processes'])
def scheduler(request):
    """Fixture to select the scheduler to use on fitness evaluation"""
    return request.param


@fixture(scope="class")
def scores(population, use_fitness):
    """Fixture to return population scores"""
    return [ph.score for ph in population]


@fixture(scope="class", autouse=True)
def use_fitness(evaluator, population):
    """Fixture to run fitness evaluation on the phenotypes population"""
    evaluator(population)


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Fitness instances attributes"""

    def test_is_callable(self, evaluator):
        """Test fitness is a callable with arity 1"""
        assert callable(evaluator)
        assert len(signature(evaluator).parameters) == 1

    def test_attr_score(self, evaluator):
        """Test score is a fitness method"""
        assert hasattr(evaluator, "score")
        assert ismethod(evaluator.score)


class ExecutionRequirements:
    """Tests group for Fitness execution features"""

    def test_score_result(self, population):
        """Test all evaluated phenotypes have a float score attr"""
        assert all(hasattr(p, "score") for p in population)
        assert all(isinstance(p.score, float) for p in population)

    @mark.parametrize("use_cache", [False], indirect=True)
    def test_no_cache(self, evaluator, population, scores):
        """Test when cache=False phenotypes are reevaluated"""
        evaluator(population)  # Run fitness a second round
        assert scores != [ph.score for ph in population]

    @mark.parametrize("use_cache", [True], indirect=True)
    def test_run_cache(self, evaluator, population, scores):
        """Test when cache=True phenotypes are not reevaluated"""
        evaluator(population)  # Run fitness a second round
        assert scores == [ph.score for ph in population]

    def test_setup_executed(self, evaluator):
        """Test set up function is executed before evaluation"""
        assert hasattr(evaluator, "executed")


# Parametrization ---------------------------------------------------
class TestFitness(AttrRequirements, ExecutionRequirements):
    """Parametrization for testing using TimeFitness"""

    @fixture(scope="class")
    def evaluator(self, fitness, use_cache, scheduler):
        """Parametrization to define the fitness method to use"""
        return fitness(cache=use_cache, scheduler=scheduler)



"""Configuration subpackage loader for evolutionary algorithms."""

import configparser
import pathlib
import os

settings = configparser.ConfigParser()
GEVOPY_USER_CONFIG = os.getenv("GEVOPY_USER_CONFIG", default="")


settings.read(pathlib.Path(__file__).parent / "defaults.toml")
settings.read(GEVOPY_USER_CONFIG)



"""Evolution algorithms with python.
The :mod:`__init__` module is intended to initialise the package and contain
the master classes to carry on the evolution. From this module, the user
can call for the `Experiment` in order to generate an instance with the
required methods and data to evolve some initial population to a desired
state.
"""

from gevopy.experiments import Experiment
from gevopy.fitness import FitnessModel
from gevopy.genetics import GenotypeModel

__all__ = [
    "Experiment",
    "FitnessModel",
    "GenotypeModel",
]



"""The :mod:`crossover` module is intended to contain abstract and specific
classes to design your evolution algoritms.

Crossover is the stage of a genetic algorithm in which individual genomes
are crossed and combined in the population for breeding.

A rework form:
https://github.com/DEAP/deap/blob/master/deap/tools/crossover.py
"""
# pylint: disable=too-few-public-methods

import builtins
from abc import ABC, abstractmethod

import numpy as np

from gevopy import genetics, utils

methods = {"OnePoint", "TwoPoint", "MultiPoint", "Uniform"}
__all__ = methods.union({"Crossover"})


class Crossover(ABC):
    """Base class for crossover methods. In genetic algorithms and
    evolutionary computation, crossover, also called recombination, is
    a genetic operator used to combine the genetic information of two
    parents to generate new offspring. It is one way to stochastically
    generate new solutions from an existing population, and is analogous
    to the crossover that happens during sexual reproduction in biology.
    """

    def __call__(self, phenotype_1, phenotype_2):
        """Executes the crossover between phenotypes.
        :param phenotype_1: First phenotype to apply crossover
        :param phenotype_2: Second phenotype to apply crossover
        :return: Crossover phenotypes
        """
        match (phenotype_1, phenotype_2):
            case _ if not isinstance(phenotype_1, genetics.GenotypeModel):
                raise ValueError("Expected 'GenotypeModel' for 'phenotype_1'")
            case _ if not isinstance(phenotype_2, genetics.GenotypeModel):
                raise ValueError("Expected 'GenotypeModel' for 'phenotype_2'")

        parents = phenotype_1, phenotype_2  # Short code
        children = tuple(phenotype.clone() for phenotype in parents)
        parents_id = [ph.id for ph in parents]
        generation = max(0, *[x.generation for x in parents]) + 1

        for child in children:  # Assign generation and parents
            child.generation = generation
            child.parents = parents_id

        if phenotype_1 != phenotype_2:  # Cross only if different parents
            self.cross_features(*[x.__dict__.values() for x in children])
        return children

    def cross_features(self, features_1, features_2):
        """Recursively crosses phenotype features. For example a list of
        diploids in the case of Eucaryote genotype.
        :param features_1: Phenotype 1 list of feature values
        :param features_2: Phenotype 2 list of feature values
        """
        for values in zip(features_1, features_2):
            if all(isinstance(v, genetics.Chromosome) for v in values):
                self.cross_chromosomes(values[0], values[1])
            elif all(isinstance(v, list) for v in values):
                self.cross_features(values[0], values[1])

    @abstractmethod
    def cross_chromosomes(self, chromosome_1, chromosome_2):
        """Executes the crossover between chromosome. The two chromosomes
        are modified in place and both keep their original length.
        :param chromosome_1: First chromosome participating in the crossover
        :param chromosome_2: Second chromosome participating in the crossover
        """
        raise NotImplementedError

    @classmethod
    def __get_validators__(cls):
        yield cls.class_validator

    @classmethod
    def class_validator(cls, value):
        """Validates the value is a correct Mutation type."""
        if not isinstance(value, cls):
            raise TypeError("'Mutation' type required")
        return value


class Uniform(Crossover):
    """Executes uniform point crossover on the input phenotypes chromosomes.
    Phenotype chromosomes are crossed on equal indexes. The resulting
    chromosomes will respectively have the length of the other.
    """

    def __init__(self, indpb=0.1):
        """Constructor for uniform point crossover.
        :param indpb: Crossover probability on each index
        """
        match type(indpb):
            case builtins.float if not 0.0 <= indpb <= 1.0:
                raise ValueError("Value for 'indpb' must be beween 0 and 1")
            case builtins.float:
                self.index_probability = indpb
            case _:
                raise ValueError("Type for 'indpb' must be 'float'")

    def cross_chromosomes(self, chromosome_1, chromosome_2):
        """Executes the crossover between chromosomes. The two chromosomes
        are modified in place and both keep their original length.
        :param chromosome_1: First chromosome participating in the crossover
        :param chromosome_2: Second chromosome participating in the crossover
        """
        mask = np.random.random(len(chromosome_1)) <= self.index_probability
        ch1, ch2 = chromosome_1, chromosome_2  # Code len reduction
        ch1[mask], ch2[mask] = ch2[mask], ch1[mask]


class MultiPoint(Crossover):
    """Executes a multiple point crossover on the input phenotypes chromosomes.
    Phenotype chromosomes are crossed on equal indexes. The resulting
    chromosomes will respectively have the length of the other.
    """

    def __init__(self, n):
        """Constructor for multiple point crossover.
        :param n: Amount of random index points to cross
        """
        match type(n):
            case builtins.int if n < 1:
                raise ValueError("Value for 'n' cannot be lower than 1")
            case builtins.int:
                self.number_points = n
            case _:
                raise ValueError("Type for 'n' must be a positive 'int'")

    def cross_chromosomes(self, chromosome_1, chromosome_2):
        """Executes the crossover between chromosomes. The two chromosomes
        are modified in place and both keep their original length.
        :param chromosome_1: First chromosome participating in the crossover
        :param chromosome_2: Second chromosome participating in the crossover
        """
        lengths = len(chromosome_1), len(chromosome_2)
        points = np.random.randint(min(*lengths), size=self.number_points)
        points.sort()
        utils.cross_chromosomes(chromosome_1, chromosome_2, points)


class OnePoint(MultiPoint):
    """Executes a one point crossover on the input phenotypes chromosomes.
    Phenotype chromosomes are crossed on equal indexes. The resulting
    chromosomes will respectively have the length of the other.
    """

    def __init__(self):
        """Constructor for one point crossover."""
        super().__init__(n=1)


class TwoPoint(MultiPoint):
    """Executes a two point crossover on the input phenotypes chromosomes.
    Phenotype chromosomes are crossed on equal indexes. The resulting
    chromosomes will respectively have the length of the other.
    """

    def __init__(self):
        """Constructor for two point crossover."""
        super().__init__(n=2)



"""The :mod:`experiments` module is intended to contain the classes
needed to execute the defined evolutionary algorithms.

The main class Experiment allows the user to create an object where
to group all the definitions in order to prepare them for execution.
It allows to create context session where phenotypes can be added and
parameters can be modified.

Session method `run` triggers the execution of the evolution process
and returns an Execution object with all the relevant statistics and
parameters at the end of the evolution process.
"""

import contextlib
import logging
import uuid
from typing import List, Optional

from pydantic import (BaseModel, Extra, Field, PositiveInt, PrivateAttr,
                      root_validator)

import gevopy.algorithms
import gevopy.database
import gevopy.fitness
import gevopy.tools
from gevopy.database import EmptyInterface
from gevopy.fitness import FitnessModel
from gevopy.genetics import GenotypeModel
from gevopy.tools import crossover, mutation, selection

# https://docs.python.org/3/howto/logging-cookbook.html
module_logger = logging.getLogger(__name__)

# Default algorithm for base experiment
DEFAULT_ALGORITHM = gevopy.algorithms.Standard(
    selection1=selection.Ponderated(),
    selection2=selection.Uniform(),
    crossover=crossover.OnePoint(),
    mutation=mutation.SinglePoint(mutpb=0.1),
    survival_rate=0.2,
)


class Experiment(BaseModel):
    """Base class for evolution experiments.
    Provides the essential attributes to create and run an experiment.
    :param fitness: Fitness instance to evaluate phenotypes
    :param algorithm: Algorithm instance to evolve phenotypes
    :param name: Experiment name, if none, generates an uuid4 string
    :param database: Database interface object, defaults to EmptyInterface
    """
    fitness: gevopy.fitness.FitnessModel
    algorithm: gevopy.algorithms.Algorithm = DEFAULT_ALGORITHM
    database: gevopy.database.Interface = EmptyInterface()
    name: str = Field(default_factory=lambda: str(uuid.uuid4()))
    _logger: logging.Logger = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._logger = logging.getLogger(f"{__package__}.Experiment")
        self._logger = self.Logger(self.logger, {"exp": self})

    class Logger(logging.LoggerAdapter):
        # pylint: disable=missing-class-docstring
        def process(self, msg, kwargs):
            return f"[{self.extra['exp'].name}]: {msg}", kwargs

    @property
    def logger(self):
        """Experiment logger, used to trace and print experiment info.
        :return: Experiment.Logger
        """
        return self._logger

    @contextlib.contextmanager
    def session(self, *args, **kwds):
        """Function to generate a context session to interface the experiment.
        :param args: Same as database.session possitional arguments
        :param kwds: Same as database.session key arguments
        """
        with self.database.session(*args, **kwds) as db_session:
            self.logger.debug("Enter session with: %s %s", args, kwds)
            yield Session(experiment=self, database=db_session)
            self.logger.debug("Exit session with: %s %s", args, kwds)

    def close(self,  *args, **kwds):
        """Function to close the database interface driver.
        :param args: Same as database Driver close possitional arguments
        :param kwds: Same as database Driver close key arguments
        """
        self.database.close(*args, **kwds)


class Session(BaseModel):
    """Base class for evolution experiment sessions.
    Provides the essential attributes to prepare the execution conditions.
    :param experiment: Experiment name the session is linked with
    :param database: Database interface session,
    """

    experiment: Experiment
    database: gevopy.database.SessionContainer
    _population: List[GenotypeModel] = PrivateAttr(default=[])

    class Config:
        # pylint: disable=missing-class-docstring
        arbitrary_types_allowed = True

    def save_phenotypes(self, phenotypes):
        """Saves the phenotypes to the experiment database.
        :param phenotypes: List of phenotypes to add to the experiment
        """
        if any(not isinstance(x, GenotypeModel) for x in phenotypes):
            raise ValueError("Phenotypes must inherit from GenotypeModel")
        iserial_phenotypes = (p.dict(serialize=True) for p in phenotypes)
        self.database.add_phenotypes(iserial_phenotypes)  # Iter to speed up

    def run(self, **execution_kwds):
        """Executes the algorithm until a stop condition is met.
        :param max_generation: The maximum number of loops to run
        :param max_score: The score required to stop the evolution
        :return: Generated Execution instance
        """
        return Execution(**execution_kwds).run(session=self)

    def add_phenotypes(self, phenotypes, save=True):
        """Adds phenotypes to the experiment session population.
        :param phenotypes: List of phenotypes to add to the experiment
        :param save: Flag to save new population status in database
        """
        for phenotype in phenotypes:  # Add experiment to phenotypes
            phenotype.experiment = self.experiment.name
        if save:
            self.save_phenotypes(phenotypes)
        self._population += phenotypes

    def get_phenotypes(self):
        """Gets population phenotypes from the experiment session.
        :return: Pool with experiment session phenotypes
        """
        return gevopy.tools.Pool(self._population)

    def del_experiment(self):
        """Deletes the experiment phenotypes and data. Also in database.
        """
        self.database.del_experiment(name=self.experiment.name)
        self._population = []

    def eval_phenotypes(self, fitness, save=True):
        """Executes the fitness evaluation on the session population.
        :param algorithm: Algorithm to run each execution generation cycle
        :param save: Flag to save new population status in database
        """
        if not isinstance(fitness, FitnessModel):
            raise ValueError("Expected 'FitnessModel' type for 'fitness'")
        fitness(self._population)
        if save:
            self.save_phenotypes(self._population)

    def generate_offspring(self, algorithm, save=True):
        """Replaces population with offspring generated from algorithm.
        :param algorithm: Algorithm to run for producing the offspring
        :param save: Flag to save new population status in database
        """
        if not isinstance(algorithm, gevopy.algorithms.Algorithm):
            raise ValueError("Expected 'Algorithm' type for 'algorithm'")
        self._population = algorithm(self._population)
        if save:
            self.save_phenotypes(self._population)


class Execution(BaseModel):
    """Base class for evolution algorithm execution. This class uses an
    experiment session to run evolution cycles and generations on a population
    of phenotypes. It also includes statistics about the execution process.

    Note that if neither max_generation or max_score are defined, the
    constructor raises ValueError for required valid end conditions.
    """
    max_generation: Optional[PositiveInt] = None
    max_score: Optional[float]
    halloffame: gevopy.tools.HallOfFame = gevopy.tools.HallOfFame(3)
    generation: int = 0

    class Config:
        # pylint: disable=missing-class-docstring
        # pylint: disable=too-few-public-methods
        extra = Extra.forbid

    def __repr__(self) -> str:
        return (
            "Evolutionary algorithm execution report:\n"
            f"  Executed generations: {self.generation}\n"
            f"  Best phenotype: {self.halloffame[0].id}\n"
            f"  Best score: {self.best_score}\n"
        )

    @root_validator()
    def check_max_gen_or_score(cls, values):
        """Checks for valid end conditions in the Execution"""
        # pylint: disable=no-self-argument
        max_generation = values.get('max_generation')
        max_score = values.get("max_score")
        if (max_generation is None) and (max_score is None):
            raise ValueError('Either max_generation or max_score is required')
        return values

    @property
    def best_score(self):
        """Best score reached by the evaluated phenotypes during the run.
        :return: Float (not only positive)
        """
        try:  # If not started, halloffame is empty and raises IndexError
            return self.halloffame[0].score
        except IndexError:  # Empty if not started
            return None

    class RunLogger(logging.LoggerAdapter):
        # pylint: disable=missing-class-docstring
        def process(self, msg, kwargs):
            return f"[gen:{self.extra['exe'].generation}]: {msg}", kwargs

    def run(self, session):
        """Executes the algorithm until a stop condition is met.
        :param session: Experiment session used for the evolution execution
        :return: Completed Execution instance (statistics)
        """
        logger = self.RunLogger(session.experiment.logger, dict(exe=self))
        algorithm = session.experiment.algorithm
        fitness = session.experiment.fitness
        try:
            logger.info("Start of evolutionary experiment execution")
            session.eval_phenotypes(fitness, save=True)  # Evaluate first pop
            while not self.completed:
                self.generation += 1  # Increase generation index
                session.generate_offspring(algorithm, save=False)
                session.eval_phenotypes(fitness, save=True)
                self.halloffame.update(session.get_phenotypes())
                logger.info("Completed cycle; %s", self.best_score)
        except Exception as err:
            logger.error("Error %s raised during experiment execution", err)
            raise err
        else:
            logger.info("Experiment execution completed successfully")
            return self

    @property
    def completed(self):
        """Evaluates if the final generation or required score is reached.
        :return: True if evolution conditions are met, False otherwise
        """
        if self.max_generation:
            if self.generation and self.max_generation <= self.generation:
                return True
        if self.max_score:
            if self.best_score and self.max_score <= self.best_score:
                return True
        return False  # If any of the defined



"""Module to test genotype requirements"""
# pylint: disable=redefined-outer-name

import uuid
from collections.abc import MutableSequence
from datetime import datetime

from pytest import fixture

from gevopy import genetics


# Module fixtures ---------------------------------------------------
@fixture(scope="class")
def clone(phenotype):
    """Fixture to return a phenotype clone"""
    return phenotype.clone()


# Requirements ------------------------------------------------------
class AttrRequirements:
    """Tests group for Genotype instances attributes"""

    def test_attr_chromosome(self, phenotype):
        """Test phenotype has a correct 'chromosome' attribute"""
        assert hasattr(phenotype, "chromosome")
        assert isinstance(phenotype.chromosome, MutableSequence)

    def test_attr_id(self, phenotype):
        """Test phenotype has a correct 'id' attribute"""
        assert hasattr(phenotype, "id")
        assert isinstance(phenotype.id, uuid.UUID)

    def test_attr_created(self, phenotype):
        """Test phenotype has a correct 'created' attribute"""
        assert hasattr(phenotype, "created")
        assert isinstance(phenotype.created, datetime)

    def test_attr_parents(self, phenotype):
        """Test phenotype has a correct 'parents' attribute"""
        assert hasattr(phenotype, "parents")
        assert isinstance(phenotype.parents, list)

    def test_attr_generation(self, phenotype):
        """Test phenotype has a correct 'generation' attribute"""
        assert hasattr(phenotype, "generation")
        assert isinstance(phenotype.generation, int)
        assert phenotype.generation > 0

    def test_attr_score(self, phenotype):
        """Test phenotype has a correct 'score' attribute"""
        assert hasattr(phenotype, "score")
        assert isinstance(phenotype.score, type(None))

    def test_is_instance_genotype(self, phenotype):
        """Test phenotype is instance of GenotypeModel"""
        assert isinstance(phenotype, genetics.GenotypeModel)


class CloneRequirements:
    """Tests group for phenotype clone function"""

    def test_clone_id(self, phenotype, clone):
        """Test clone 'id' is different from phenotype"""
        assert clone.id != phenotype.id

    def test_clone_chromosome(self, phenotype, clone):
        """Test clone 'chromosome' is equal to phenotype"""
        assert clone.chromosome == phenotype.chromosome

    def test_clone_parents(self, phenotype, clone):
        """Test clone 'parents' are equal to phenotype"""
        assert clone.parents == phenotype.parents

    def test_clone_generation(self, phenotype, clone):
        """Test clone 'generation' is equal to phenotype"""
        assert clone.generation == phenotype.generation


# Parametrization ---------------------------------------------------
class TestGenotype(AttrRequirements, CloneRequirements):
    """Parametrization for testing Genotypes"""

    @fixture(scope="class")
    def phenotype(self, genotype):
        """Fixture to return a genotype instance"""
        return genotype()



"""Examples module for users on how to create fitness functions and testing"""
import numpy as np

from gevopy.fitness import FitnessModel


# ------------------------------------------------------------------
# Most Ones --------------------------------------------------------
# This fitness object scores each phenotypes with the amount of ones
# the phenotype has in it's chromosome attribute.
class MostOnes(FitnessModel):
    """Fitness model count amount of '1' in the chromosome"""

    def score(self, phenotype):
        return phenotype.chromosome.count(1)


# ------------------------------------------------------------------
# Fitness ----------------------------------------------------------
# This fitness object scores each phenotypes completelly random.
class Random(FitnessModel):
    """Fitness model assigns a random score between 0-1"""
    # pylint: disable=attribute-defined-outside-init

    def setup(self, phenotypes):
        """Set up method to run once per generation"""
        self.executed = True

    def score(self, phenotype):
        """Method to use for evaluation phenotypes"""
        return np.random.random()
